@@ -10,12 +10,18 @@ import re
 import subprocess
 import sys
 import time
+import warnings
 from collections import defaultdict
 
 import six
 import pkg_resources
 import requests
 
+try:
+    import argcomplete
+except ImportError:
+    argcomplete = None
+
 from copr import CoprClient
 import copr.exceptions as copr_exceptions
 from copr.v3 import (
@@ -49,6 +55,14 @@ ON_OFF_MAP = {
     None: None,
 }
 
+BOOTSTRAP_MAP = {
+    "default": "default",
+    "on": "on",
+    "off": "off",
+    "image": "image",
+    None: "default",
+}
+
 no_config_warning = """
 ================= WARNING: =======================
 File '{0}' is missing or incorrect.
@@ -81,6 +95,32 @@ class FrontendOutdatedCliException(Exception):
         )
 
 
+class ActionDeprecated(argparse.Action):
+    """ automate deprecation warnings for options """
+    def __call__(self, parser, namespace, values, option_string=None):
+        options = ", ".join(self.option_strings)
+        warnings.warn("Use of DEPRECATED option: {0}".format(options))
+        # keep the normal "store" behavior
+        setattr(namespace, self.dest, values)
+
+
+def buildopts_from_args(args, progress_callback):
+    """
+    For all the build commands, parse the common set of build options.
+    """
+    buildopts = {
+        "timeout": args.timeout,
+        "chroots": args.chroots,
+        "background": args.background,
+        "progress_callback": progress_callback,
+    }
+    for opt in ["exclude_chroots", "bootstrap", "after_build_id", "with_build_id", "isolation"]:
+        value = getattr(args, opt)
+        if value is not None:
+            buildopts[opt] = value
+    return buildopts
+
+
 class Commands(object):
     def __init__(self, config_path):
         self.config_path = config_path or '~/.config/copr'
@@ -140,6 +180,17 @@ class Commands(object):
             owner = self.config["username"]
         return owner, name
 
+    def parse_chroot_path(self, path):
+        """
+        Take a `path` in an `owner/project/chroot` format and return a tuple of
+        the corresponding values, i.e. `(owner, project, chroot)`
+        """
+        m = re.match(r"(([^/]+)/)?([^/]+)/(.*)", path)
+        if m:
+            owner = m.group(2) or self.config["username"]
+            return owner, m.group(3), m.group(4)
+        raise CoprException("Unexpected chroot path format")
+
     def _watch_builds(self, build_ids):
         """
         :param build_ids: list of build IDs
@@ -291,6 +342,18 @@ class Commands(object):
         }
         return self.process_build(args, self.client.build_proxy.create_from_scm, data)
 
+    @requires_api_auth
+    def action_build_distgit_simple(self, args):
+        """ build-distgit method """
+        data = {
+            "packagename": args.pkgname,
+            "distgit": args.instance,
+            "namespace": args.namespace,
+            "committish": args.committish,
+        }
+        return self.process_build(
+            args, self.client.build_proxy.create_from_distgit, data)
+
     @requires_api_auth
     def action_build_rubygems(self, args):
         """
@@ -331,9 +394,7 @@ class Commands(object):
         projectname = project_dirname.split(':')[0]
 
         try:
-            buildopts = {"memory": args.memory, "timeout": args.timeout, "chroots": args.chroots,
-                         "background": args.background, "progress_callback": progress_callback}
-
+            buildopts = buildopts_from_args(args, progress_callback)
             result = build_function(ownername=username, projectname=projectname,
                                     project_dirname=project_dirname, buildopts=buildopts, **data)
 
@@ -376,7 +437,8 @@ class Commands(object):
             enable_net=ON_OFF_MAP[args.enable_net],
             persistent=args.persistent,
             auto_prune=ON_OFF_MAP[args.auto_prune],
-            use_bootstrap_container=ON_OFF_MAP[args.use_bootstrap_container],
+            bootstrap=BOOTSTRAP_MAP[args.bootstrap],
+            isolation=args.isolation,
             delete_after_days=args.delete_after_days,
             multilib=ON_OFF_MAP[args.multilib],
             module_hotfixes=ON_OFF_MAP[args.module_hotfixes],
@@ -398,7 +460,8 @@ class Commands(object):
             unlisted_on_hp=ON_OFF_MAP[args.unlisted_on_hp],
             enable_net=ON_OFF_MAP[args.enable_net],
             auto_prune=ON_OFF_MAP[args.auto_prune],
-            use_bootstrap_container=ON_OFF_MAP[args.use_bootstrap_container],
+            bootstrap=BOOTSTRAP_MAP[args.bootstrap],
+            isolation=args.isolation,
             chroots=args.chroots,
             delete_after_days=args.delete_after_days,
             multilib=ON_OFF_MAP[args.multilib],
@@ -551,11 +614,16 @@ class Commands(object):
 
         :param args: argparse arguments provided by the user
         """
-        owner, copr, chroot = parse_chroot_path(args.coprchroot)
-        project_chroot = self.client.project_chroot_proxy.edit(
+
+        if args.bootstrap_image:
+            args.bootstrap = 'image'
+        owner, copr, chroot = self.parse_chroot_path(args.coprchroot)
+        self.client.project_chroot_proxy.edit(
             ownername=owner, projectname=copr, chrootname=chroot,
             comps=args.upload_comps, delete_comps=args.delete_comps,
-            additional_packages=args.packages, additional_repos=args.repos
+            additional_packages=args.packages, additional_repos=args.repos,
+            bootstrap=args.bootstrap, bootstrap_image=args.bootstrap_image,
+            isolation=args.isolation,
         )
         print("Edit chroot operation was successful.")
 
@@ -565,7 +633,7 @@ class Commands(object):
 
         :param args: argparse arguments provided by the user
         """
-        owner, copr, chroot = parse_chroot_path(args.coprchroot)
+        owner, copr, chroot = self.parse_chroot_path(args.coprchroot)
         project_chroot = self.client.project_chroot_proxy.get(
             ownername=owner, projectname=copr, chrootname=chroot
         )
@@ -626,6 +694,26 @@ class Commands(object):
             package = self.client.package_proxy.edit(ownername, projectname, args.name, "scm", data)
         print("Create or edit operation was successful.")
 
+    @requires_api_auth
+    def action_add_or_edit_package_distgit(self, args):
+        """ process 'add/edit-package-distgit' requests """
+        data = {
+            # package name (args.name) is not needed in data
+            "distgit": args.instance,
+            "namespace": args.namespace,
+            "committish": args.committish,
+            "max_builds": args.max_builds,
+            "webhook_rebuild": ON_OFF_MAP[args.webhook_rebuild],
+        }
+        ownername, projectname = self.parse_name(args.copr)
+        if args.create:
+            self.client.package_proxy.add(ownername, projectname, args.name,
+                                          "distgit", data)
+        else:
+            self.client.package_proxy.edit(ownername, projectname, args.name,
+                                           "distgit", data)
+        print("Create or edit operation was successful.")
+
     @requires_api_auth
     def action_add_or_edit_package_rubygems(self, args):
         ownername, projectname = self.parse_name(args.copr)
@@ -719,13 +807,7 @@ class Commands(object):
     def action_build_package(self, args):
         ownername, project_dirname = self.parse_name(args.copr_repo)
         projectname = project_dirname.split(':')[0]
-
-        buildopts = {
-            "chroots": args.chroots,
-            #"memory": args.memory,
-            "timeout": args.timeout,
-            "background": args.background,
-        }
+        buildopts = buildopts_from_args(args, None)
         try:
             build = self.client.package_proxy.build(ownername=ownername, projectname=projectname,
                                                     packagename=args.name, buildopts=buildopts,
@@ -745,8 +827,11 @@ class Commands(object):
         ownername, projectname = self.parse_name(args.copr)
 
         if args.yaml:
-            module = self.client.module_proxy.build_from_file(
-                    ownername, projectname, args.yaml, distgit=args.distgit)
+            try:
+                module = self.client.module_proxy.build_from_file(
+                        ownername, projectname, args.yaml, distgit=args.distgit)
+            except FileNotFoundError as e:
+                raise CoprRequestException("File '{filename}' not found".format(filename=e.filename))
         else:
             module = self.client.module_proxy.build_from_url(
                     ownername, projectname, args.url, distgit=args.distgit)
@@ -879,8 +964,22 @@ def setup_parser():
     parser_create.add_argument("--auto-prune", choices=["on", "off"], default="on",
                                help="If auto-deletion of project's obsoleted builds should be enabled (default is on).\
                                This option can only be specified by a COPR admin.")
-    parser_create.add_argument("--use-bootstrap", choices=["on", "off"], dest="use_bootstrap_container",
-                               help="If mock bootstrap container is used to initialize the buildroot.")
+    parser_create.add_argument("--isolation", choices=["simple", "nspawn", "default"], default="default",
+                               help="Choose the isolation method for running commands in buildroot.")
+
+    parser_create.add_argument(
+        "--bootstrap",
+        choices=["default", "on", "off", "image"],
+        help=(
+            "Configure Mock's bootstrap feature (consult 'man mock' for more "
+            "info).  'on'/'off' enables/disables bootstrap.  The 'default' "
+            "variant uses pre-configured setup from mock-core-configs.  The "
+            "'image' variant enforces the bootstrap initialization from "
+            "the pre-configured container image (defined in "
+            "mock-core-configs.rpm)."
+        ),
+    )
+
     parser_create.add_argument("--delete-after-days", default=None, metavar='DAYS',
                                help="Delete the project after the specfied period of time")
     parser_create.add_argument("--module-hotfixes", choices=["on", "off"], default="off",
@@ -916,8 +1015,15 @@ def setup_parser():
     parser_modify.add_argument("--auto-prune", choices=["on", "off"],
                                help="If auto-deletion of project's obsoleted builds should be enabled.\
                                This option can only be specified by a COPR admin.")
-    parser_modify.add_argument("--use-bootstrap", choices=["on", "off"], dest="use_bootstrap_container",
-                               help="If mock bootstrap container is used to initialize the buildroot.")
+    parser_modify.add_argument("--isolation", choices=["simple", "nspawn", "default"],
+                               help="Choose the isolation method for running commands in buildroot.")
+
+    parser_modify.add_argument(
+        "--bootstrap",
+        choices=["default", "on", "off", "image"],
+        help=("Configure Mock's bootstrap feature, "
+              "See 'create --help' for more info."))
+
     parser_modify.add_argument("--delete-after-days", default=None, metavar='DAYS',
                                help=("Delete the project after the specfied "
                                      "period of time, empty or -1 disables, "
@@ -982,6 +1088,27 @@ def setup_parser():
                                         choices=["rpkg", "tito", "tito_test", "make_srpm"],
                                         help="Srpm build method. Default is 'rpkg'.")
 
+    parser_distgit_simple_parent = argparse.ArgumentParser(add_help=False)
+    parser_distgit_simple_parent.add_argument(
+        "--commit", dest="committish", default=None,
+        help="Branch name, tag name, or git hash to built the package from")
+    parser_distgit_simple_parent.add_argument(
+        "--namespace", dest="namespace", default=None,
+        help=(
+            "Some DistGit instances (e.g. the Fedora Copr dist-git) use "
+            "a namespaced clone/lookaside URLs.  Typically it meas that "
+            "one package may be hosted in the same DistGit instance "
+            "multiple times, in multiple namespaces.  Specify the NAMESPACE "
+            "here (e.g. @copr/copr for @copr/copr/copr-cli package)."),
+    )
+    parser_distgit_simple_parent.add_argument(
+        "--distgit", dest="instance", default=None,
+        help=(
+            "Dist-git instance to build the package from, for example "
+            "'fedora'."
+        ),
+    )
+
     parser_rubygems_args_parent = argparse.ArgumentParser(add_help=False)
     parser_rubygems_args_parent.add_argument("--gem", metavar="GEM", dest="gem_name",
                                              help="Specify gem name")
@@ -1019,21 +1146,58 @@ def setup_parser():
                                      help="The copr repo to build the package in. Can be just name of project or even in format username/project or @groupname/project. "
                                      "It can also be in the form project:<tag>, which will put the build into a side repository with the user-chosen tag in its name.")
     parser_build_parent.add_argument("--memory", dest="memory",
-                                     help="")
-    parser_build_parent.add_argument("--timeout", dest="timeout",
-                                     help="")
+                                     help=argparse.SUPPRESS,
+                                     action=ActionDeprecated)
+    parser_build_parent.add_argument(
+        "--timeout", dest="timeout",
+        help=("Specify build timeout (seconds), if the build takes "
+              "longer than that, it is terminated and fails.  The "
+              "default is usually set to 5 hours on Copr Frontend."))
     parser_build_parent.add_argument("--nowait", action="store_true", default=False,
                                      help="Don't wait for build")
     parser_build_parent.add_argument("-r", "--chroot", dest="chroots", action="append",
                                      help="If you don't need this build for all the project's chroots. You can use it several times for each chroot you need.")
+
+    parser_build_parent.add_argument(
+        "--exclude-chroot",
+        dest="exclude_chroots",
+        action="append",
+        help=("If you don't need this build for all the project's chroots."
+              "You can use it several times for each chroot you don't need.")
+    )
+
     parser_build_parent.add_argument("--background", dest="background", action="store_true", default=False,
                                      help="Mark the build as a background job. It will have lesser priority than regular builds.")
+    parser_build_parent.add_argument("--isolation", choices=["simple", "nspawn", "default"], default="unchanged",
+                                     help="Choose the isolation method for running commands in buildroot.")
+
+    parser_build_parent.add_argument(
+        "--bootstrap",
+        choices=["unchanged", "default", "on", "off", "image"],
+        help=("Configure Mock's bootstrap feature, "
+              "default is 'unchanged' so the configuration from Copr project "
+              "and Copr chroot is used for this build. "
+              "The 'default' variant resets the project/chroot configuration "
+              "to the pre-configured setup from mock-core-configs. "
+              "See 'create --help' for more info."))
+
+    batch_build_opts = parser_build_parent.add_mutually_exclusive_group()
+    batch_build_opts.add_argument(
+        "--after-build-id", metavar="BUILD_ID",
+        help=("Build after the batch containing the BUILD_ID build."),
+    )
+
+    batch_build_opts.add_argument(
+        "--with-build-id", metavar="BUILD_ID",
+        help=("Build in the same batch with the BUILD_ID build."),
+    )
 
     # create the parser for the "build" (url/upload) command
     parser_build = subparsers.add_parser("build", parents=[parser_build_parent],
                                          help="Build packages to a specified copr")
     parser_build.add_argument("pkgs", nargs="+",
                               help="filename of SRPM or URL of packages to build")
+
     parser_build.set_defaults(func="action_build")
 
     # create the parser for the "buildpypi" command
@@ -1063,6 +1227,22 @@ def setup_parser():
                                               help="Builds package from Git/DistGit/SVN repository.")
     parser_build_scm.set_defaults(func="action_build_scm")
 
+    # create the parser for the "build-distgit" command
+    parser_build_distgit_simple = subparsers.add_parser(
+        "build-distgit",
+        parents=[parser_distgit_simple_parent, parser_build_parent],
+        help="Builds a package from a DistGit repository",
+        description=(
+            "Build a package from a DistGit repository. "
+            "For more info about DistGit build method see the description "
+            "'add-package-distgit' command."),
+    )
+    parser_build_distgit_simple.add_argument(
+        "--name", dest="pkgname", required=True,
+        help=("Package name to build from the DistGit instance"),
+    )
+    parser_build_distgit_simple.set_defaults(func="action_build_distgit_simple")
+
     # create the parser for the "status" command
     parser_status = subparsers.add_parser("status", help="Get build status of build specified by its ID")
     parser_status.add_argument("build_id", help="Build ID", type=int)
@@ -1112,6 +1292,24 @@ def setup_parser():
                                       help="space separated string of package names to be added to buildroot")
     parser_edit_chroot.add_argument("--repos",
                                       help="space separated string of additional repo urls for chroot")
+    parser_edit_chroot.add_argument("--isolation", choices=["simple", "nspawn", "default"], default="unchanged",
+                                    help="Choose the isolation method for running commands in buildroot.")
+
+    parser_edit_chroot.add_argument(
+        "--bootstrap",
+        choices=["unchanged", "default", "on", "off", "image"],
+        help=("Configure Mock's bootstrap feature, "
+              "default is 'unchanged' so the configuration from Copr project "
+              "is used for this chroot. "
+              "The 'default' variant resets the project configuration "
+              "to the pre-configured setup from mock-core-configs. "
+              "See 'create --help' for more info."))
+
+    parser_edit_chroot.add_argument(
+        "--bootstrap-image",
+        help=("Use a custom container image for initializing Mock's "
+              "bootstrap (Implies --bootstrap=image)"))
+
     parser_edit_chroot.set_defaults(func="action_edit_chroot")
 
     parser_get_chroot = subparsers.add_parser("get-chroot", help="Get chroot of a project")
@@ -1164,6 +1362,43 @@ def setup_parser():
                                                         parents=[parser_scm_args_parent, parser_add_or_edit_package_parent])
     parser_edit_package_scm.set_defaults(func="action_add_or_edit_package_scm", create=False)
 
+    # DistGit edit/create package
+    parser_add_package_distgit = subparsers.add_parser(
+        "add-package-distgit",
+        help="Creates a new DistGit package",
+        description=(
+            "DistGit (Distribution Git) is Git with additional data "
+            "storage (so called \"lookaside cache\"). It is designed to hold "
+            "content of source RPMs. For more info, see the "
+            "https://github.com/release-engineering/dist-git documentation."
+            "\n\n"
+            "To build a package for a particular distribution, you need "
+            "clone the correct git repository, and download corresponding "
+            "sources files from the lookaside cache.  Each distribution though "
+            "uses a different hostname for DistGit server and may store the "
+            "git repositories and source files on a little bit different "
+            "URIs (or even in NAMESPACEs).  That's why Copr has this "
+            "build method pre-configured as \"DistGit instances\" (one "
+            "instance per one distribution)."
+        ),
+        parents=[parser_distgit_simple_parent,
+                 parser_add_or_edit_package_parent])
+    parser_add_package_distgit.set_defaults(
+        func="action_add_or_edit_package_distgit",
+        create=True)
+    parser_edit_package_distgit = subparsers.add_parser(
+        "edit-package-distgit",
+        help="Edits an existing DistGit package",
+        description=(
+            "Edit an existing DistGit package.  For more info about DistGit "
+            "build method see the description of 'add-package-distgit' "
+            "command."),
+        parents=[parser_distgit_simple_parent,
+                 parser_add_or_edit_package_parent])
+    parser_edit_package_distgit.set_defaults(
+        func="action_add_or_edit_package_distgit",
+        create=False)
+
     # Rubygems edit/create
     parser_add_package_rubygems = subparsers.add_parser("add-package-rubygems",
                                                         help="Creates a new RubyGems package",
@@ -1353,16 +1588,11 @@ def setup_parser():
             help=request_help.format('builder'))
     parser_permissions_request.set_defaults(func='action_permissions_request')
 
+    if argcomplete:
+        argcomplete.autocomplete(parser)
     return parser
 
 
-def parse_chroot_path(path):
-    m = re.match(r"(([^/]+)/)?([^/]+)/(.*)", path)
-    if m:
-        return m.group(2), m.group(3), m.group(4)
-    return None
-
-
 def enable_debug():
     logging.basicConfig(
         level=logging.DEBUG,
