@@ -47,7 +47,7 @@ else:
 
 def serializable(result):
     if isinstance(result, dict):
-        new_result = result.copy()
+        new_result = dict(result)
         new_result.pop("__response__", None)
         new_result.pop("__proxy__", None)
         return new_result
